package de.qwertyuiop.copret
import os.Path

import Terminal.*
import syntax.*

case class Presentation(
    slides: Vector[Slide],
    meta: Map[String, String] = Map.empty,
):
  def start(using keymap: Keymap = Keymap.default) =
    Terminal.enterRawMode()
    Terminal.hideCursor()
    Thread.sleep(1000)
    run()

  import Presentation._
  def run()(using keymap: Keymap) =
    import SlideAction.*
    @annotation.tailrec
    def rec(pos: Int, action: SlideAction): Unit =

      inline def redraw() = rec(pos - 1, QuickNext)

      inline def navigate(pos: Int, condition: Boolean, direction: Int)(
          executor: Int => Slide => Unit,
      ) =
        if condition then
          executor(pos + direction)(slides(pos))
          rec(pos + direction, waitkey)
        else rec(pos, waitkey)

      inline def runInteractive(cmd: Vector[String], path: Path) =
        Terminal.showCursor()
        try os.proc(cmd).call(cwd = path)
        catch case _ => ()
        Terminal.hideCursor()
        Terminal.clear()
        redraw()

      action match
        case Start                  => navigate(0, true, 0)(executeSlide)
        case Next                   => navigate(pos, pos + 1 < slides.size, 1)(executeSlide)
        case QuickNext              => navigate(pos, pos + 1 < slides.size, 1)(executeQuick)
        case Prev                   => navigate(pos, pos > 0, -1)(executeQuick)
        case Interactive(cmd, path) => runInteractive(cmd, path)

        case Goto(target) =>
          for i <- 0 until target
          do executeSilent(i)()
          rec(target - 1, QuickNext)

        case GotoSelect =>
          promptSlide() match
            case Some(i) => rec(pos, Goto(i))
            case None    => redraw()

        case Help =>
          Terminal.clear()
          println(keymap.help)
          waitkey
          redraw()

        case Other(codes) =>
          Terminal.printStatus(action.show)
          rec(pos, waitkey)

        case Quit =>
          Terminal.showCursor()
      end match
    end rec
    rec(0, Start)
  end run

  def promptSlide() =
    val maxSlide = slides.size - 1
    prompt(s"Go to slide (0 - $maxSlide):", _.toIntOption)(
      (res, input) => res.filter((0 to maxSlide).contains).isEmpty && input.nonEmpty,
      in => s"No such slide: $in (empty input to abort)",
    )

  def executeSlide(pos: Int)(
      slide: Slide = slides(pos),
  ): Unit = slide match
    case Paragraph(contents)                        => println(contents)
    case Clear                                      => Terminal.clear()
    case PauseKey                                   => waitkey(using Keymap.empty)
    case Pause(msec)                                => Thread.sleep(msec)
    case incMd @ IncludeMarkdown(_)                 => println(incMd.markdownBlock())
    case Image(file, None)                          => println(KittyGraphicsProtocol.showImage(file))
    case Image(file, Some(ImageSize(w, h, aspect))) =>
      println(KittyGraphicsProtocol.showImage(file)) // TODO
    case cmd: TypedCommand[_]                       => cmd.show()
    case Silent(actions)                            => actions()
    case Group(slides)                              => slides.foreach(executeSlide(pos))
    case lios @ LazyIOSlide(_, display)             => executeSlide(pos)(lios.genSlide())
    case Meta(genSlide)                             => executeSlide(pos)(genSlide(this, pos))

  def executeQuick(pos: Int)(
      slide: Slide = slides(pos),
  ): Unit = slide match
    case Pause(msec)                    => ()
    case PauseKey                       => ()
    case cmd: TypedCommand[_]           => cmd.quickShow()
    case Group(slides)                  => slides.foreach(executeQuick(pos))
    case lios @ LazyIOSlide(_, display) => executeQuick(pos)(lios.genSlide())
    case _                              => executeSlide(pos)(slide)

  def executeSilent(pos: Int)(
      slide: Slide = slides(pos),
  ): Unit = slide match
    case cmd: TypedCommand[_]                                              => cmd.force()
    case Group(slides)                                                     => slides.foreach(executeSilent(pos))
    case lios @ LazyIOSlide(_, display)                                    =>
      executeSilent(pos)(lios.genSlide())
    case Paragraph(_) | Image(_, _) | Clear | IncludeMarkdown(_) | Meta(_) => ()
    case _                                                                 => executeQuick(pos)(slide)
end Presentation

case class ImageSize(width: Double, height: Double, keepAspect: Boolean)

sealed trait Slide
case class Paragraph(contents: String) extends Slide:
  def centerVertical(height: Int): Paragraph =
    val lines = contents.toString.count(_ == '\n') + 1
    val pad   = "\n" * ((height - lines) / 2)
    Paragraph(pad + contents + pad)

object Paragraph:
  def apply(str: fansi.Str): Paragraph = Paragraph(str.toString)

case class IncludeMarkdown(path: Path) extends Slide:
  def markdownBlock() =
    %%%(
      "/usr/bin/mdcat",
      "--columns",
      (columns * 0.8).toInt.toString,
      path.toString,
    )(using os.pwd).block

case class Image(path: Path, sizing: Option[ImageSize]) extends Slide
object Image:
  def apply(path: Path)                                                      = new Image(path, None)
  def scaled(path: Path, width: Double, height: Double, keepAspect: Boolean) =
    Image(path, Some(ImageSize(width, height, keepAspect)))

case object Clear                                       extends Slide
case class Pause(millisec: Long)                        extends Slide
case object PauseKey                                    extends Slide
case class Meta(contents: (Presentation, Int) => Slide) extends Slide

case class TypedCommand[T](
    exec: T => String,
    display: String,
    cmd: T,
    cmdIsHidden: Boolean,
    outputIsHidden: Boolean,
) extends Slide:
  private lazy val _output = exec(cmd)
  def output               = _output

  infix def showing(s: String): TypedCommand[T] = TypedCommand(exec, s, cmd)

  def show() =
    force()
    if !cmdIsHidden then
      prompt()
      Terminal.showCursor()
      typeCmd()
    if !outputIsHidden then print(output)
    if !cmdIsHidden then Terminal.hideCursor()

  def quickShow() =
    force()
    if !cmdIsHidden then
      prompt()
      println(display)
    if !outputIsHidden then print(output)

  def prompt() = print(fansi.Color.LightGreen("user@host % "))
  def force()  = _output

  def display(s: String) = copy(display = s)

  private def typeCmd() =
    for char <- display do
      print(char)
      Thread.sleep(50 + scala.util.Random.nextInt(80))
    println()

  /* Conditionally disable execution. Useful for e.g. a debug mode, or a non-interactive mode */
  def disable(altDisplay: String = display, condition: Boolean = true) =
    if condition then copy(display = altDisplay, exec = (_: T) => "")
    else this

  /* Conditionally replace the executed command (but still displaying the same). Useful for e.g. a non-interactive mode,
   * where a call to an editor is replaced with a file operation */
  def replaceIf(condition: Boolean)(tc: TypedCommand[_]): TypedCommand[_] =
    if condition then tc.showing(display)
    else this
end TypedCommand

object TypedCommand:
  val shell = sys.env.getOrElse("SHELL", "sh")

  def run(using Path): Vector[String] => String =
    c => runProcess(c)

  def runShell(using Path): Vector[String] => String =
    c => runProcess(Vector(shell, "-c", c.mkString(" ")))

  def runInteractive(using Path): Vector[String] => String =
    c => { os.proc(c).call(); "" }

  def apply(cmd: String*)(using Path): TypedCommand[Vector[String]] =
    TypedCommand(run, cmd.mkString(" "), cmd.toVector)

  def apply[T](exec: T => String, display: String, cmd: T) =
    new TypedCommand(exec, display, cmd, false, false)

  def shell(cmd: String*)(using Path): TypedCommand[Vector[String]] =
    TypedCommand(runShell, cmd.mkString(" "), cmd.toVector)

  def fake(cmd: String): TypedCommand[String] =
    TypedCommand(_ => "", cmd, cmd)

  def interactive(cmd: String*)(using Path): TypedCommand[Vector[String]] =
    TypedCommand(runInteractive, cmd.mkString(" "), cmd.toVector)

  def untyped(cmd: String*)(using Path): TypedCommand[Vector[String]] =
    TypedCommand(exec = run, display = "", cmd.toVector, cmdIsHidden = true, outputIsHidden = false)
end TypedCommand

sealed abstract case class Silent[T] private (doStuff: () => T) extends Slide
object Silent:
  def apply[T](doStuff: => T) = new Silent(() => doStuff) {}

case class Group(slides: List[Slide]) extends Slide
object Group:
  def apply(slides: Slide*): Group = Group(slides.toList)

case class LazyIOSlide[T](runOnce: () => T, display: T => Slide) extends Slide:
  private lazy val data = runOnce()
  def genSlide(): Slide = display(data)

trait SlideSyntax:
  private[copret] class LazyIOSlideBuilder[T](runOnce: => T):
    def useIn(display: T => Slide) = LazyIOSlide(() => runOnce, display)

  def prepare[T](runOnce: => T): LazyIOSlideBuilder[T] = LazyIOSlideBuilder(
    runOnce,
  )

/* vim:set tw=120: */
